/*
 * Copyright 2003, 2004 Porchdog Software. All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or without modification,
 *	are permitted provided that the following conditions are met:
 *
 *		1. Redistributions of source code must retain the above copyright notice,
 *		   this list of conditions and the following disclaimer.   
 *		2. Redistributions in binary form must reproduce the above copyright notice,
 *		   this list of conditions and the following disclaimer in the documentation
 *		   and/or other materials provided with the distribution.
 *
 *	THIS SOFTWARE IS PROVIDED BY PORCHDOG SOFTWARE ``AS IS'' AND ANY
 *	EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *	IN NO EVENT SHALL THE HOWL PROJECT OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 *	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *	BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 *	OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *	OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 *	OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	The views and conclusions contained in the software and documentation are those
 *	of the authors and should not be interpreted as representing official policies,
 *	either expressed or implied, of Porchdog Software.
 */

#include "posix_interface.h"
#include <salt/vstdlib.h>
#include <salt/vstring.h>
#include <salt/verrno.h>
#include <salt/assert.h>
#include <salt/log.h>
#include <salt/interface.h>
#include <net/if.h> 
#include <arpa/inet.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <stdio.h>

#if defined(__linux__)
typedef unsigned long long u64;
typedef __uint32_t 	u32;
typedef __uint16_t 	u16;
typedef __uint8_t		u8;
#	include <linux/sockios.h>
#	include <linux/ethtool.h> 
#endif

#ifndef SIOCGIFCONF
#	include <sys/sockio.h>
#endif

#define SW_E_BAD_ARG 23

#define SW_TRY_GOTO_ERROR(EXPR)  { if ((result = EXPR) != SW_OKAY) goto error; } ((void) 0)

sw_result
sw_network_interface_link_status(
			sw_network_interface 	self, 
			sw_bool					*	islinked);

sw_result
sw_network_interfaces_fina(
	sw_ulong 					nifc,
	sw_network_interface * 	nifv);

static sw_result
sw_posix_inet_socket(int * fd);


/*
	sockaddr_dl is only referenced if we're using IP_RECVIF, 
   so only include the header in that case.
*/

#ifdef  IP_RECVIF
#	include <net/if_dl.h>
#endif

#define max(a,b)	((a) > (b) ? (a) : (b))
/* for Posix and MacOs */
#if defined(__APPLE__) || defined(__FreeBSD__) || defined(__NetBSD__)
int 
get_sa_len(struct ifreq * ifr)
{
	return max(sizeof(struct sockaddr), ifr->ifr_addr.sa_len);
}
#endif
/* for Posix */
#if defined(__linux__)
int 
get_sa_len(struct ifreq * ifr)
{
#ifdef  IPV6
	if (ifr->ifr_addr.sa_family == AF_INET6)
	{
		return sizeof(struct sockaddr_in6);
	}
#endif
	/* AF_INET */
	return sizeof(struct sockaddr);
}
#endif  

#if defined(__sun)
int
get_sa_len(struct ifreq * ifr)
{
	return sizeof(struct sockaddr);
}
#endif

static sw_result
sw_posix_inet_socket(int * fd)
{
	if ((*fd = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
	{
		sw_log(SW_LOG_ERROR, "sw_posix_inet_socket", "socket failed: %s\n", strerror(errno));
		return SW_E_INIT;
	}

	return SW_OKAY;
}

sw_string 
sw_mac_address_name(
				sw_mac_address	self,
				sw_string		name,
				sw_ulong			len)
{
	sprintf(name, "%.2x:%.2x:%.2x:%.2x:%.2x:%.2x", self.m_id[0], self.m_id[1], self.m_id[2], self.m_id[3], self.m_id[4], self.m_id[5]);

	return name;
}

sw_result 
sw_network_interface_init(
            sw_network_interface   *  netif)
{
	sw_posix_network_interface	pnetif;
	sw_result 						result = SW_OKAY;

   /* allocate a posix network interface */
   SW_TRY(sw_malloc(pnetif, sw_posix_network_interface, sizeof(struct _sw_posix_network_interface)));

	/* netif points to the first field we allocated */
	*netif = &pnetif->m_super;

   return result;
}

sw_result 
sw_network_interface_fina(
			sw_network_interface self)
{
	sw_posix_network_interface pnetif = (sw_posix_network_interface) self;
	
	sw_free(pnetif);

	return SW_OKAY;
}

sw_result
sw_network_interface_exists(
		sw_string name, 
		sw_bool * exists)
{
#if defined(__linux__)
	sw_result 		result = SW_OKAY;
	int 				s;
	struct ifreq	req;
	int				r;

	/* get a socket for ioctling */
	SW_TRY(sw_posix_inet_socket(&s));
	
	memset(&req, 0, sizeof(req));
	strncpy(req.ifr_name, name, IFNAMSIZ);

	if ((r = ioctl(s, SIOCGIFINDEX, &req)) < 0 && errno != ENODEV) 
	{
		sw_log(SW_LOG_ERROR, "sw_network_interface_exists: %s\n", strerror(errno));
		result = SW_E_INIT;
	}

	*exists = (r >= 0 && req.ifr_ifindex >= 0);

	close(s);

	return result;
#else
	return SW_E_FAIL;
#endif
}

sw_result 
sw_network_interface_up(
					sw_network_interface self)
{
	sw_result		result = SW_E_INIT;
	int 				fd;
	struct ifreq 	ifr;

	/* get a socket for ioctling */
	SW_TRY(sw_posix_inet_socket(&fd));

	memset(&ifr, 0, sizeof(ifr));
	strncpy(ifr.ifr_name, self->m_name, sizeof(ifr.ifr_name)-1);

	/* get the interface flags */
	if (ioctl(fd, SIOCGIFFLAGS, &ifr) < 0) 
	{
		sw_log(SW_LOG_WARNING, "sw_network_interface_up", "Warning: Could not get interface flags.\n", strerror(errno));

		goto exit;
	}

	/* if it's already up, we're exit */
	if ((ifr.ifr_flags & IFF_UP) == IFF_UP)
	{
		result = SW_OKAY;
		goto exit;
	}

	/* get interface flags */
	if (ioctl(fd, SIOCGIFFLAGS, &ifr) < 0) 
	{
		sw_log(SW_LOG_WARNING, "sw_network_interface_up", "Warning: Could not get interface flags.\n", strerror(errno));
		goto exit;
	}

	/* bring the interface up */
	ifr.ifr_flags |= IFF_UP;

	/* set interface flags */
	if (ioctl(fd, SIOCSIFFLAGS, &ifr) < 0)
	{
		sw_log(SW_LOG_WARNING, "sw_network_interface_up", "Warning: Could not set interface flags.\n", strerror(errno));
		goto exit;
	}

	result = SW_OKAY;

exit:
	
	close(fd);

	return result;
}


sw_result 
sw_network_interface_link_status(
			sw_network_interface 	self, 
			sw_bool					*	islinked)
{
#if defined(__linux__)
	sw_result 				result = SW_OKAY;
	int 						fd;
	struct ifreq			ifr;
	int						r;
	struct ethtool_value edata;

	sw_assert(self != NULL);		/* must be initialized */
	sw_assert(islinked != NULL);	/* must be allocated */

	/* get a socket for ioctling */
	SW_TRY(sw_posix_inet_socket(&fd));

	memset(&ifr, 0, sizeof(ifr));
	strncpy(ifr.ifr_name, self->m_name, sizeof(ifr.ifr_name) - 1);

	edata.cmd = ETHTOOL_GLINK;
	ifr.ifr_data = (caddr_t) &edata;

	if (ioctl(fd, SIOCETHTOOL, &ifr) < 0)
	{
		/*
		 * if we fail, we'll limp on
		 */
		sw_log(SW_LOG_WARNING, "sw_network_interface_link_status", "ETHTOOL_GLINK failed: %s\n", strerror(errno));

		/*
		 * pretend we're linked
		 */
		edata.data = 1;
	}

	*islinked = edata.data ? SW_TRUE : SW_FALSE;

exit:

	close(fd);

	return result;
#else
	return SW_OKAY;
#endif
}


sw_result 
sw_network_interface_set_ipv4_address(
				sw_network_interface		self,
				sw_ipv4_address			addr)
{
	return sw_ipv4_address_init_from_address(&self->m_ipv4_address, addr);
}

static sw_result
sw_posix_network_interface_init_from_name(
	sw_posix_network_interface	nif,
	sw_string						ifname)
{
	sw_result		result = SW_E_INIT;
	int 				sock;
	char			*	cptr;
	sw_bool			islinked;
	struct ifreq	ifr;
	char				tmpname[18];

	sw_assert(nif != NULL);			/* already allocated */
	sw_assert(ifname != NULL);		/* already initialized */

	/* make sure there won't be any problems */
	sw_memset(nif, 0, sizeof(struct _sw_posix_network_interface));
	if (sw_strlen(ifname) > IFNAMSIZ)
	{
		ifname[IFNAMSIZ - 1] = (char)0;
	}

	/* name */
	sw_strcpy(nif->m_super.m_name, ifname);

	/* get a socket for ioctling */
	SW_TRY(sw_posix_inet_socket(&sock));

	strcpy(ifr.ifr_name, ifname);
	ifr.ifr_addr.sa_family = AF_INET;

	/* ip address */
	if ((ioctl(sock, SIOCGIFADDR, &ifr)) < 0)
	{
		sw_saddr addrAny = 0;

		sw_debug("couldn't get ip address for %s, setting ip to 0\n", ifr.ifr_name);
		SW_TRY_GOTO(sw_ipv4_address_init_from_saddr(
			&(nif->m_super.m_ipv4_address), addrAny));
	}
	else
	{
		SW_TRY_GOTO(sw_ipv4_address_init_from_saddr(
			&(nif->m_super.m_ipv4_address), 
			((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr.s_addr));
	}
	sw_ipv4_address_name(nif->m_super.m_ipv4_address, tmpname, 18);
	sw_debug("got ip address: %s\n", tmpname);

	/* mac address */
#if defined(__linux__)
	if ((ioctl(sock, SIOCGIFHWADDR, &ifr)) < 0)
	{
		sw_log(SW_LOG_ERROR, "sw_posix_network_interface_init_from_name", "unable to get mac address\n");
		goto exit;
	}
	sw_memcpy(nif->m_super.m_mac_address.m_id, (sw_octet*)(ifr.ifr_hwaddr.sa_data), sizeof(sw_mac_address)); 
#endif

	/* ifnet flags */
#if defined(__linux__)
	if (ioctl(sock, SIOCGIFFLAGS, &ifr) < 0)
	{
		sw_log(SW_LOG_ERROR, "sw_posix_network_interface_init_from_name", "unable to get interface flags");
		goto exit;
	}
#endif
	  
	/* index */
#if defined(__linux__)
	if (ioctl(sock, SIOCGIFINDEX, &ifr) < 0)
	{
		sw_log(SW_LOG_ERROR, "sw_posix_network_interface_init_from_name", "unable to get interface flags");
		goto exit;
	}
	nif->m_super.m_index = ifr.ifr_ifindex;
#endif

	/* initialize link status field */
	SW_TRY_GOTO(sw_network_interface_link_status(&nif->m_super, &(nif->m_super.m_linked)));

	result = SW_OKAY;

exit:

	close(sock);

	return result;
}

static sw_result
sw_posix_network_interface_init_from_ifreq(
	sw_posix_network_interface	nif,
	struct ifreq				*	ifr)
{
#if defined(__linux__)
	sw_result	result = SW_E_INIT;
	int 			sock;
	char		*	cptr;
	sw_bool		islinked;

	sw_assert(nif != NULL);		/* already allocated */
	sw_assert(ifr != NULL);		/* already initialized */

	/* get a socket for ioctling */
	SW_TRY(sw_posix_inet_socket(&sock));

	/* ip address */
	sw_ipv4_address_init_from_saddr(&(nif->m_super.m_ipv4_address), ((struct sockaddr_in *) &ifr->ifr_addr)->sin_addr.s_addr);

	/* mac address */
	if ((ioctl(sock, SIOCGIFHWADDR, ifr)) < 0)
	{
		sw_log(SW_LOG_ERROR, "sw_posix_network_interface_init_from_ifreq", "unable to get mac address");
		goto exit;
	}
	sw_memcpy(nif->m_super.m_mac_address.m_id, (sw_octet*)(ifr->ifr_hwaddr.sa_data), sizeof(sw_mac_address)); 

	/* ifnet flags */
	if (ioctl(sock, SIOCGIFFLAGS, ifr) < 0)
	{
		sw_log(SW_LOG_ERROR, "sw_posix_network_interface_init_from_ifreq", "unable to get interface flags");
		goto exit;
	}
	  
	/* name */
	if ( (cptr = strchr(ifr->ifr_name, ':')) != NULL)
	{
		*cptr = 0;      
	}
	
	sw_memcpy(nif->m_super.m_name, ifr->ifr_name, IFNAMSIZ);
	nif->m_super.m_name[IFNAMSIZ-1] = '\0';

	/* index */
	nif->m_super.m_index = ifr->ifr_ifindex;

	/* initialize link status field */
	if ((result = sw_network_interface_link_status(&nif->m_super, &islinked)) != SW_OKAY)
	{
		sw_log(SW_LOG_ERROR, "sw_posix_network_interface_init_from_ifreq", "sw_network_interface_link_status failed\n");
		goto exit;
	}
	nif->m_super.m_linked = islinked;

	result = SW_OKAY;

exit:

	close(sock);

	return result;
#else
	return SW_E_FAIL;
#endif
}

static sw_result
get_iflist_buffer(int sock, struct ifconf * ifc)
{
	int 	lastlen = 0;

	sw_assert(ifc != NULL);

	/* we'll guess at the buflen until we grab enough */
	ifc->ifc_len = 10 * sizeof(struct ifreq);
	for ( ; ; )
	{
		/* allocate buffer big enough for if config data */
   	SW_TRY(sw_malloc(ifc->ifc_buf, char*, ifc->ifc_len));

		/* get ifnet config */
		if (ioctl(sock, SIOCGIFCONF, ifc) < 0)
		{
			if (errno != EINVAL || lastlen != 0)
			{
				/* not enough memory */
				sw_free(ifc->ifc_buf);
				return SW_E_INIT;
			}
		}
		else
		{			
			/* if we get the same size twice, we're exit */
			if (ifc->ifc_len == lastlen)
			{
				return SW_OKAY;   /* success */
			}
				
			lastlen = ifc->ifc_len;
		}
		  
		/* bump up the size */
		ifc->ifc_len += 10 * sizeof(struct ifreq);   

		/* free the last guess */
		sw_free(ifc->ifc_buf);
	}
}

#define _PATH_PROCNET_DEV "/proc/net/dev"

static int
procenetdev_version(char * buf)
{
	if (strstr(buf, "compressed"))
		return 3;
	if (strstr(buf, "bytes"))
		return 2;

	return 1;
}

static char * 
get_name(char * name, char *p)
{
    while (isspace(*p))
   p++;
    while (*p) {
   if (isspace(*p))
       break;
   if (*p == ':') {  /* could be an alias */
       char *dot = p, *dotname = name;
       *name++ = *p++;
       while (isdigit(*p))
      *name++ = *p++;
       if (*p != ':') { /* it wasn't, backup */
      p = dot;
      name = dotname;
       }
       if (*p == '\0')
      return NULL;
       p++;
       break;
   }
   *name++ = *p++;
    }
    *name++ = '\0';
    return p;

}

sw_result
sw_network_interfaces2(
		sw_ulong					* 	nifc,
		sw_network_interface	**	nifv)
{
	FILE * fh;
	char buf[512];
	int procnetdev_vsn;
	sw_result result = SW_OKAY;

	if ((fh = fopen(_PATH_PROCNET_DEV, "r")) == NULL)
	{
		sw_log(SW_LOG_ERROR, "sw_network_interfaces2", "cannot open %s (%s).\n", _PATH_PROCNET_DEV, strerror(errno));
		goto exit;
	}

	fgets(buf, sizeof buf, fh); /* eat line */
	fgets(buf, sizeof buf, fh); /* eat line */

	/* TODO hard code 10 for now */
	/* allocate nifv  */
   SW_TRY(sw_malloc(*nifv, sw_network_interface*, 10 * sizeof(sw_network_interface)));
	
	*nifc = 0;
	while (fgets(buf, sizeof(buf), fh))
	{
		char * s, name[IFNAMSIZ];
		sw_posix_network_interface nif;
		sw_ipv4_address ipaddr;

		s = get_name(name, buf);

		if (sw_strcmp(name, "lo") == 0)
		{
			continue;
		}

		/* create a new netif */
		SW_TRY_GOTO_ERROR(sw_network_interface_init((sw_network_interface*)&nif)); 

		/* initialize fields */
		SW_TRY_GOTO_ERROR(sw_posix_network_interface_init_from_name(nif, name));

		SW_TRY_GOTO_ERROR(sw_network_interface_ipv4_address((sw_network_interface)nif, &ipaddr));

		(*nifv)[(*nifc)++] = (sw_network_interface)nif;
	}

	if (ferror(fh))
	{
		perror(_PATH_PROCNET_DEV);
		result = SW_E_INIT;
		goto error;
	}

	result = SW_OKAY;

	goto exit;

error:

	if (*nifv != NULL)
	{
		sw_network_interfaces_fina(*nifc, *nifv);
	}

exit:

	if (fh != NULL)
	{
		fclose(fh);
	}

	return result;
}

sw_result
sw_network_interfaces(
		sw_ulong					* 	nifc,
		sw_network_interface	**	nifv)
{
	sw_result		result = SW_E_INIT;
	int         	sock;
	struct ifconf 	ifc;
	struct ifreq*	ifr;
	char 			*	ptr;
	int         	i = 0, n = 0, estimate = 0;
	sw_ulong			lb_saddr = sw_ipv4_address_saddr(sw_ipv4_address_loopback());

	*nifc = 0;
	ifc.ifc_buf = NULL;

	/* get a socket for ioctling */
	SW_TRY(sw_posix_inet_socket(&sock));

	SW_TRY_GOTO_ERROR(get_iflist_buffer(sock, &ifc));

	/* count netifs */
	estimate = ifc.ifc_len / sizeof(struct ifreq);

	/* allocate nifv  */
   SW_TRY(sw_malloc(*nifv, sw_network_interface*, estimate * sizeof(sw_network_interface)));

	/* creating netifs */
	ifr = ifc.ifc_req;
	for (ptr = ifc.ifc_buf, i = 0; ptr < ifc.ifc_buf + ifc.ifc_len; i < *nifc)
	{
		struct ifreq		*	ifr;
		sw_network_interface nif;
		sw_ulong					ip_saddr;
		int						salen;

		ifr = (struct ifreq *) ptr;
		salen = get_sa_len(ifr);
		ptr += sizeof(ifr->ifr_name) + salen;

		ip_saddr = ((struct sockaddr_in *) &ifr->ifr_addr)->sin_addr.s_addr;

		if (	(ifr->ifr_addr.sa_family != AF_INET) ||
				(ip_saddr == lb_saddr)	)
		{
			continue;   /* ignore loopback+non-inet */
		}

		/* create a new netif */
		SW_TRY_GOTO_ERROR(sw_network_interface_init(&nif)); 

		/* initialize fields */
		SW_TRY_GOTO_ERROR(sw_posix_network_interface_init_from_name((sw_posix_network_interface)nif, ifr->ifr_name));

		(*nifv)[(*nifc)++] = nif;
	}
    
	result = SW_OKAY;
	goto exit;
    
error:

	if (*nifv > 0)
	{
		sw_network_interfaces_fina(*nifc, *nifv);
	}

exit:
    
	if (ifc.ifc_buf != NULL)
	{
		sw_free(ifc.ifc_buf);
	}
	
	if (sock != -1)
	{
		int bfd = close(sock);
		sw_assert(bfd == 0);
	}
	
	return result;    
}

sw_result
sw_network_interfaces_fina(
	sw_ulong 					nifc,
	sw_network_interface * 	nifv)
{
	int i;

	for (i = 0; i < nifc; i++)
	{
		SW_TRY(sw_network_interface_fina(nifv[i]));
	}

	sw_free(nifv);

	return SW_OKAY;
}

sw_result 
sw_network_interface_by_name(
            sw_string               name,
            sw_network_interface *  netif)
{
	sw_result	result;

	SW_TRY(sw_network_interface_init(netif));

	if ((result = sw_posix_network_interface_init_from_name((sw_posix_network_interface)*netif, name)) != SW_OKAY)
	{
		sw_network_interface_fina(*netif);
	}

	return result;
}
