/*
 * Copyright 2003, 2004 Porchdog Software. All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or without modification,
 *	are permitted provided that the following conditions are met:
 *
 *		1. Redistributions of source code must retain the above copyright notice,
 *		   this list of conditions and the following disclaimer.   
 *		2. Redistributions in binary form must reproduce the above copyright notice,
 *		   this list of conditions and the following disclaimer in the documentation
 *		   and/or other materials provided with the distribution.
 *
 *	THIS SOFTWARE IS PROVIDED BY PORCHDOG SOFTWARE ``AS IS'' AND ANY
 *	EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *	IN NO EVENT SHALL THE HOWL PROJECT OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 *	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *	BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 *	OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *	OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 *	OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	The views and conclusions contained in the software and documentation are those
 *	of the authors and should not be interpreted as representing official policies,
 *	either expressed or implied, of Porchdog Software.
 */
#include <stdio.h>
#include <fcntl.h>
#include <nifd.h>
#include <time.h>
#include <sys/signal.h>
#include <netinet/in.h>
#include <salt/vstdlib.h>
#include <salt/verrno.h>
#include <salt/assert.h>
#include <salt/interface.h>
#include <salt/signal.h>

/* 
 * globals
 */
static sw_bool restart_interfaces = SW_TRUE;

/* 
 * Callback functions 
 */
static sw_result
timer_event_handler(
		sw_timer_handler     handler,
		sw_salt           	salt,
		sw_timer          	timer,
		sw_time              timeout,
		sw_opaque            extra);


static sw_result
sighup_command(
			sw_signal_handler	handler,
         sw_salt        	salt,
         sw_signal      	signal,
         sw_opaque      	extra);


static sw_result
sigusr1_command(
			sw_signal_handler	handler,
         sw_salt        	salt,
         sw_signal      	signal,
         sw_opaque      	extra);


static sw_result
sigusr2_command(
			sw_signal_handler	handler,
         sw_salt        	salt,
         sw_signal      	signal,
         sw_opaque      	extra);


static sw_result
sigint_command(
			sw_signal_handler	handler,
         sw_salt        	salt,
         sw_signal      	signal,
         sw_opaque      	extra);


/*
 * nifd methods
 */
static sw_result
nifd_write_pidfile(nifd * self, char * dname)
{
	sw_result 	result = SW_E_INIT;
	char 			pidfile[64];
	char 			str[16];

	umask(022);
	sprintf(pidfile, "/var/run/%s.pid", dname);

	/* open the file */
	if ((self->m_pidfd = open(pidfile, O_RDWR|O_CREAT, 0644)) < 0)
	{
		sw_log(SW_LOG_ERROR, "nifd_write_pidfile", "Can't open %s.\n", pidfile);
		goto exit;
	}
			
	/* lock it */
	if (lockf(self->m_pidfd, F_TLOCK,0) != 0)
	{
		sw_log(SW_LOG_ERROR, "nifd_write_pidfile", "Can't lock %s.  Is nifd already running?\n", pidfile);
		goto exit;
	}
	
	/* write our pid */
	sprintf(str, "%d\n", getpid());
	write(self->m_pidfd, str, strlen(str)); // record pid to lockfile

	result = SW_OKAY;

exit:

	return result;
}


static sw_result
nifd_read_pidfile(nifd self, char * dname, int * pid)
{
	sw_result 	result = SW_E_INIT;
	char 			pidfile[64];
	FILE 		*	fp;

	sprintf(pidfile, "/var/run/%s.pid", dname);
	fp = fopen(pidfile, "rt");

	if (fp == NULL || fscanf(fp, "%d", pid) < 1)
	{
		sw_log(SW_LOG_ERROR, "nifd", "%s daemon is not running\n", dname);
		goto exit;
	}

	fclose(fp);

	result = SW_OKAY;

exit:

	return result;
}

/* 
 * nifd_init()
 *
 *	initialize resources
 */
sw_result
nifd_init(nifd * self, sw_ulong interval)
{
	self->m_salt					= NULL;
	self->m_pidfd					= -1;
	self->m_timer 					= NULL;
	self->m_sigusr1 				= NULL;
	self->m_sigusr2 				= NULL;
	self->m_sighup 				= NULL;
	self->m_sigint 				= NULL;
	self->m_nifcount 				= 0;
	self->m_nifs 					= NULL;
	self->m_polltime.m_secs		= interval;
	self->m_polltime.m_usecs	= 0;

	/* create the pid file */
	SW_TRY(nifd_write_pidfile(self, "nifd"));

	/* initialize salt */
	SW_TRY(sw_salt_init(&(self->m_salt), 0, NULL));

	/* get all available interfaces */
	SW_TRY(sw_network_interfaces2(&self->m_nifcount, &self->m_nifs));

	/* initialize the timer */
	SW_TRY(sw_timer_init(&(self->m_timer)));

	/* register the timer */
	sw_debug("nifd: installing timer ...\n");
	SW_TRY(sw_salt_register_timer(
						self->m_salt, 
						self->m_timer, 
						self->m_polltime, 
						(sw_timer_handler) self, 
						timer_event_handler, 
						NULL));

	/* register signal handlers */
	sw_debug("nifd: registering signal handlers ...\n");
	SW_TRY(sw_signal_init(&self->m_sigint, SIGINT));
	SW_TRY(sw_salt_register_signal(
						self->m_salt,
						self->m_sigint,
						(sw_signal_handler) self,
						sigint_command,
						NULL));

	SW_TRY(sw_signal_init(&self->m_sigusr1, SIGUSR1));
	SW_TRY(sw_salt_register_signal(
						self->m_salt,
						self->m_sigusr1,
						(sw_signal_handler) self,
						sigusr1_command,
						NULL));

	SW_TRY(sw_signal_init(&self->m_sigusr2, SIGUSR2));
	SW_TRY(sw_salt_register_signal(
						self->m_salt,
						self->m_sigusr2,
						(sw_signal_handler) self,
						sigusr2_command,
						NULL));

	SW_TRY(sw_signal_init(&self->m_sighup, SIGHUP));
	SW_TRY(sw_salt_register_signal(
						self->m_salt,
						self->m_sighup,
						(sw_signal_handler) self,
						sighup_command,
						NULL));

	return SW_OKAY;
}

/* 
 * nifd_fina()
 *
 *	free resources
 */
sw_result
nifd_fina(nifd * self)
{
	/* free signals */
	if (self->m_sigusr1 != NULL)
	{
		SW_TRY(sw_salt_unregister_signal(self->m_salt, self->m_sigusr1));
		SW_TRY(sw_signal_fina(self->m_sigusr1));
	}

	if (self->m_sigusr2 != NULL)
	{
		SW_TRY(sw_salt_unregister_signal(self->m_salt, self->m_sigusr2));
		SW_TRY(sw_signal_fina(self->m_sigusr2));
	}

	if (self->m_sigint != NULL)
	{
		SW_TRY(sw_salt_unregister_signal(self->m_salt, self->m_sigint));
		SW_TRY(sw_signal_fina(self->m_sigint));
	}

	if (self->m_sighup != NULL)
	{
		SW_TRY(sw_salt_unregister_signal(self->m_salt, self->m_sighup));
		SW_TRY(sw_signal_fina(self->m_sighup));
	}

	/* free available interfaces */
	if (self->m_nifs != NULL)
	{
		SW_TRY(sw_network_interfaces_fina(self->m_nifcount, self->m_nifs));
	}

	/* close the pid file */
	if (self->m_pidfd != -1)
	{
		close(self->m_pidfd);
	}

	/* free timer */
	if (self->m_timer != NULL)
	{
		SW_TRY(sw_timer_fina(self->m_timer));
	}

	/* free salt */
	if (self->m_salt != NULL)
	{
		SW_TRY(sw_salt_fina(self->m_salt));
	}

	return SW_OKAY;
}

static sw_result
nifd_signal_daemon(nifd self, sw_string dname)
{
	int	sysres;
	int 	pid;

	SW_TRY(nifd_read_pidfile(self, dname, &pid));

	sw_debug("sending kill SIGUSR1 to %d\n", pid);
	if ((sysres = kill(pid, SIGUSR1)) != 0)
	{
		sw_log(SW_LOG_WARNING, "nifd_signal_ip_change", "failed to signal %s\n", dname);
	}

	return (sysres == 0) ? SW_OKAY : SW_E_INIT;
}


static sw_result
nifd_signal_autoipd(nifd self, sw_network_interface nif)
{
	char name[32];
	char dname[32];

	/* signal autoipd */
	sw_network_interface_name(nif, name, 32);
	sprintf(dname, "autoipd-%s", name);

	return nifd_signal_daemon(self, dname);
}

/* 
 * nifd_restart_interface()
 * 
 * call ifdown and ifup on link 
 */
static sw_result
nifd_restart_interface(
	nifd 						self, 
	sw_network_interface nif)
{
	int	sysres;
	char 	ifname[32];
	char	dname[32];
	char 	cmd[128];
	char*	cmds[2] = { "ifdown", "ifup" };
	int	i;

	sw_network_interface_name(nif, ifname, 32);

	for (i = 0; i < 2; i++)
	{
		sprintf(cmd, "/sbin/%s %s", cmds[i], ifname);
		if ((sysres = system(cmd)) < 0)
		{
			sw_log(SW_LOG_ERROR, "nifd_restart_interface", "failed to call /sbin/%s on %s\n", cmds[i], ifname);
		}
		else if (sysres > 0)
		{
			sw_log(SW_LOG_WARNING, "nifd_restart_interface", "/sbin/%s failed on %s\n", cmds[i], ifname);
		}
	}

	return (sysres == 0) ? SW_OKAY : SW_E_INIT;
}

/*
 * timer_event_handler()
 * 	called where the main loop detects that a timeout has occurred.
 */
static sw_result
timer_event_handler(
         sw_timer_handler  handler,
         sw_salt           salt,
         sw_timer       	timer,
         sw_time           timeout,
         sw_opaque         extra)
{
	sw_result 					result;
   nifd						*  ifmd;
	sw_ulong						newnifc;
	sw_network_interface *	newnifs;
	sw_network_interface * 	tnifs;
	sw_bool						signal_mDNS = SW_FALSE;
	int 							i;

   ifmd = (nifd*) handler;

	sw_assert(ifmd != NULL);
   sw_assert(ifmd->m_timer == timer);

	/* reinstall one-shot timer                     */
	/* NOTE: do this first, otherwise we'll hang in */
	/* 	   select as a result of any error        */
	SW_TRY(sw_salt_register_timer(
				salt, 
				ifmd->m_timer, 
				ifmd->m_polltime, 
				(sw_timer_handler) ifmd, 
				timer_event_handler, NULL));

	sw_debug("***** polling for network events ******\n");
	SW_TRY(sw_network_interfaces2(&newnifc, &newnifs));

	/* iterate through the niftable and compare states */
	for (i = 0; i < ifmd->m_nifcount; i++)
	{
		int 						j;
		sw_ulong					inifi;
		sw_network_interface	inif = ifmd->m_nifs[i];
		sw_bool					found = SW_FALSE;
		char 						ifname[32];

		sw_network_interface_name(inif, ifname, 32);

		/* get the nif index for i */
		SW_TRY_GOTO(sw_network_interface_index(inif, &inifi));

		for (j = 0; j < newnifc && !found; j++)
		{
			sw_ulong					jnifi;
			sw_bool					islinked, waslinked;
			sw_ipv4_address		newip, oldip;
			sw_network_interface	jnif = newnifs[i];

			sw_debug("\t checking link status on %s\n", ifname);

			/* get the nif index for j */
			SW_TRY_GOTO(sw_network_interface_index(jnif, &jnifi));
			
			if (jnifi == inifi)
			{
				found = SW_TRUE;

				/* check current link status */
				SW_TRY_GOTO(sw_network_interface_linked(jnif, &islinked));
				/* check previous link status */
				SW_TRY_GOTO(sw_network_interface_linked(inif, &waslinked));
				/* check current ip address */
				SW_TRY_GOTO(sw_network_interface_ipv4_address(jnif, &newip));
				/* check previous ip address */
				SW_TRY_GOTO(sw_network_interface_ipv4_address(inif, &oldip));

				/* see if link status or ip address changed */
				if (islinked != waslinked)	
				{
					sw_debug("link status changed to %s\n", (islinked) ? "up" : "down");
					if (islinked && restart_interfaces)
					{
						if ((nifd_restart_interface(*ifmd, inif)) != SW_OKAY)
						{
							sw_log(SW_LOG_WARNING, "nifd", "can't restart interface\n", ifname);
						}
					}
					else
					{
						/* nothing to do */
					}

					sw_debug("signalling autoipd\n");
					nifd_signal_autoipd(*ifmd, inif);
					signal_mDNS = SW_TRUE;
				}

				if (!sw_ipv4_address_equals(newip, oldip))
				{
					char oldname[18];
					char newname[18];

					sw_debug("ip address changed from %s to %s\n", 
						sw_ipv4_address_name(oldip, oldname, 18), 
						sw_ipv4_address_name(newip, newname, 18));
					nifd_signal_autoipd(*ifmd, inif);
					signal_mDNS = SW_TRUE;
				}
			}
		}

		/* if we don't find the nif, it's been dropped */
		if (!found)
		{
			sw_assert(SW_FALSE);

			nifd_signal_autoipd(*ifmd, inif);
			signal_mDNS = SW_TRUE;
		}
	}

	/* if we added any netifs, signal daemons */
	if (signal_mDNS || (newnifc != ifmd->m_nifcount))
	{
		sw_debug("sending SIGUSR1 to mDNSResponder\n");
		nifd_signal_daemon(*ifmd, "mDNSResponder");
	}

exit:

	/* free the old nifs */
	SW_TRY(sw_network_interfaces_fina(ifmd->m_nifcount, ifmd->m_nifs));

	/* keep the new nifs */
	ifmd->m_nifcount 	= newnifc;
	ifmd->m_nifs		= newnifs;

   return SW_OKAY;
}

static sw_result
nifd_refresh_timer(nifd * self)
{
	sw_salt_unregister_timer(self->m_salt, self->m_timer);

	SW_TRY(sw_salt_register_timer(
				self->m_salt, 
				self->m_timer, 
				self->m_polltime, 
				(sw_timer_handler) self, 
				timer_event_handler, NULL));
}

static sw_result
sighup_command(
				sw_signal_handler	handler,
            sw_salt           salt,
            sw_signal         signal,
            sw_opaque         extra)
{
	sw_assert(handler != NULL);

	return nifd_refresh_timer((nifd*)handler);
}


static sw_result
sigusr1_command(
				sw_signal_handler	handler,
            sw_salt           salt,
            sw_signal         signal,
            sw_opaque         extra)
{
	sw_assert(handler != NULL);

	return nifd_refresh_timer((nifd*)handler);
}


static sw_result
sigusr2_command(
				sw_signal_handler	handler,
            sw_salt           salt,
            sw_signal         signal,
            sw_opaque         extra)
{
	return sw_debug_memory_inuse();
}


static sw_result
sigint_command(
				sw_signal_handler	handler,
            sw_salt           salt,
            sw_signal         signal,
            sw_opaque         extra)
{
	sw_assert(handler != NULL);
	sw_assert(salt != NULL);

   return sw_salt_stop_run(salt);
}


static void
print_usage(const char * program)
{
	fprintf(stderr, "usage: %s [options]\n\n", program);
	fprintf(stderr, "Options:\n\n");
	fprintf(stderr, "-h            this help\n");
	fprintf(stderr, "-v            display version information\n");
	fprintf(stderr, "-d            run in debug mode\n");
	fprintf(stderr, "-n            don't run ifdown/ifup on relink\n");
	fprintf(stderr, "-i interval   poll interval in seconds\n");
	exit(0);
}


int
main(int argc, char ** argv)
{
	sw_result		result;
	struct _nifd ifmd;
	int				opt;
	sw_ulong			interval = 1;
	sw_bool			make_daemon = SW_TRUE;

	/* sw_debug_enable(SW_TRUE); */

	while ((opt = getopt(argc, argv, "?hvdni:")) != EOF)
	{
		switch (opt)
		{
			case '?':
			case 'h':
			{
				print_usage(argv[0]);
			}
			break;

			case 'i':
			{
				interval = atol(optarg);

				if (interval == 0)
				{
					interval = 1;
				}
			}
			break;

			case 'v':
			{
				fprintf(stderr, "nifd release " VERSION " Copyright(c) Porchdog Software.\n");
				exit(0);
			}
			break;

			case 'd':
			{
				make_daemon = SW_FALSE;
			}
			break;

			case 'n':
			{
				restart_interfaces = SW_FALSE;
			}
			break;

			default:
			{
				print_usage(argv[0]);
			}
			break;
		}
	}

	if (getuid() != 0)
	{
		fprintf(stderr, "nifd must be run as root\n");
		exit(1);
	}

	if (make_daemon)
	{
		if (daemon(SW_FALSE, SW_FALSE) < 0)
		{
			sw_log(SW_LOG_FATAL_ERROR, "main", "unable to daemonize\n");
		}
	}

	fprintf(stderr, "im here\n");

	SW_TRY(nifd_init(&ifmd, interval));

	fprintf(stderr, "nifd starting up...\n");
	SW_TRY_GOTO(sw_salt_run(ifmd.m_salt));
	fprintf(stderr, "nifd shutting down...\n");

exit:

	SW_TRY(nifd_fina(&ifmd));

	return SW_OKAY;
}
